"""Module with the main constants from Kytos.

These constants may be overridden by values passed on the controller
instantiation.
"""
CONNECTION_TIMEOUT = 70
# FLOOD_TIMEOUT in microseconds
FLOOD_TIMEOUT = 100000
