"""Holds main metadata information about the project.

The present metadata is intended to be used mainly on the setup.
"""
__version__ = '2020.1'
__author__ = 'Kytos Team'
__author_email__ = 'devel@lists.kytos.io'
__license__ = 'MIT'
__url__ = 'https://github.com/kytos/kytos'
__description__ = 'Kytos, an open source SDN Platform'
